from functools import reduce
from typing import Union, Optional, Hashable

import attr
import funcy as fn


@attr.s(auto_attribs=True, frozen=True)
class Context:
    node_val: Union[str, bool]
    negated: bool
    max_lvl: int
    node: Hashable
    path_negated: bool
    curr_lvl: Optional[int] = None
    prev_lvl: Optional[int] = None
    low_lvl: Optional[int] = None
    high_lvl: Optional[int] = None

    @property
    def is_leaf(self):
        return self.curr_lvl == self.max_lvl

    @property
    def skipped_decisions(self):
        if self.prev_lvl is None:
            return self.curr_lvl
        return self.curr_lvl - self.prev_lvl - 1

    @property
    def skipped_paths(self):
        return 2**self.skipped_decisions


def _ctx(node, manager, prev_ctx=None):
    max_lvl = len(manager.vars)
    path_negated = False if prev_ctx is None else \
        (prev_ctx.path_negated ^ prev_ctx.negated)
    common = {
        "node": node,
        "negated": node.negated,
        "path_negated": path_negated,
        "prev_lvl": None if prev_ctx is None else prev_ctx.curr_lvl,
        "max_lvl": max_lvl,
        "curr_lvl": min(node.level, max_lvl)
    }

    if node.var is None:
        specific = {
            "node_val": node == manager.true,
        }
    else:
        specific = {
            "node_val": node.var,
            "low_lvl": min(node.low.level, max_lvl),
            "high_lvl": min(node.high.level, max_lvl),
        }

    return Context(**common, **specific)


def post_order(node, merge, *, manager=None, prev_ctx=None):
    if manager is None:
        manager = node.bdd

    @fn.memoize
    def _post_order(node, prev_ctx=None):
        ctx = _ctx(node, manager, prev_ctx=prev_ctx)

        if ctx.is_leaf:
            return merge(ctx=ctx, low=None, high=None)

        def _reduce(c):
            return _post_order(c, prev_ctx=ctx)

        return merge(ctx=ctx, high=_reduce(node.high), low=_reduce(node.low))

    return _post_order(node)


def path(node, vals):
    vals = list(vals)
    prev_lvl, offset = node.level, 0
    while node.var is not None:
        prev_lvl = node.level
        if len(vals) > 1:
            val, *vals = vals[offset:]
        else:
            assert len(vals) > 0
            val = vals[0]

        yield node, val
        node = node.high if val else node.low
        offset = node.level - prev_lvl - 1
        assert offset >= 0

    yield node, None


def fold_path(merge, bexpr, vals, initial=None, manager=None):
    if manager is None:
        manager = bexpr.bdd

    def acc(prev_ctx_acc, node_val):
        prev_ctx, acc = prev_ctx_acc
        node, val = node_val
        ctx = _ctx(node, manager, prev_ctx=prev_ctx)
        return (ctx, merge(ctx, val, acc))

    return reduce(acc, path(bexpr, vals), (None, initial))[1]



import funcy as fn
from itertools import combinations_with_replacement as combinations
from itertools import product

try:
    from dd.cudd import BDD
except ImportError:
    from dd.autoref import BDD

from fold_bdd import post_order, fold_path
from fold_bdd.folds import path


def create_manager():
    manager = BDD()
    manager.declare('x', 'y')
    manager.reorder({'x': 1, 'y': 0})
    manager.configure(reordering=False)
    return manager


def test_post_order():
    @fn.memoize
    def merge1(ctx, low=None, high=None):
        return 1 if low is None else low + high

    manager = create_manager()
    bexpr = manager.add_expr('x & y')

    val = post_order(bexpr, merge1)
    assert val == bexpr.dag_size

    # Convert BDD to function.
    def merge2(ctx, low=None, high=None):
        if ctx.is_leaf:
            return lambda _: ctx.node_val

        def _eval(vals):
            val, *vals2 = vals
            out = high(vals2) if val else low(vals2)
            return ctx.negated ^ out

        return _eval

    _eval = post_order(bexpr, merge2)
    for vals in combinations([True, False], 2):
        assert all(vals) == _eval(vals)

    def merge3(ctx, low, high):
        if ctx.is_leaf:
            return ctx.skipped_paths if ctx.node_val else 0

        return (low + high) * ctx.skipped_paths

    bexpr2 = manager.add_expr('x | y')

    assert post_order(bexpr, merge3) == 1
    assert post_order(bexpr2, merge3) == 3


def test_path():
    manager = create_manager()
    bexpr = manager.add_expr('x | y')

    questions = {
        (False, False): 3,
        (False, True): 3,
        (True, False): 2,
        (True, True): 2,
    }

    for q, a in questions.items():
        assert len(list(path(bexpr, q))) == a


def test_fold_path():
    manager = create_manager()
    bexpr = manager.add_expr('x | y')
    bexpr2 = manager.true

    def merge(ctx, val, acc):
        return acc + 1

    def count_nodes(bexpr, vals):
        return fold_path(merge, bexpr, vals, initial=0)

    assert count_nodes(bexpr, (False, False)) == 3
    assert count_nodes(bexpr, (False, True)) == 3
    assert count_nodes(bexpr, (True, False)) == 2
    assert count_nodes(bexpr, (True, True)) == 2

    assert count_nodes(bexpr2, (True, True)) == 1
    assert count_nodes(bexpr2, (False, True)) == 1

    def merge2(ctx, val, acc):
        return acc * ctx.skipped_paths

    def count_paths(bexpr, vals):
        return fold_path(merge2, bexpr, vals, initial=1)

    assert count_paths(bexpr, (False, False)) == 1
    assert count_paths(bexpr, (False, True)) == 1
    assert count_paths(bexpr, (True, False)) == 2
    assert count_paths(bexpr, (True, True)) == 2

    assert count_paths(bexpr2, (True, True)) == 4
    assert count_paths(bexpr2, (False, True)) == 4


def test_path_negation():
    levels = {'c1': 0, 'a1': 1, 'c2': 2, 'a2': 3}
    manager = BDD(levels)
    manager.configure(reordering=False)

    c1, a1, c2, a2 = map(manager.var, ['c1', 'a1', 'c2', 'a2'])
    bexpr = ((c1 & a1) | (~c1 & ~a1)) & ((c2 & a2) | (~c2 & ~a2))

    assert bexpr.low.negated
    assert not bexpr.high.negated

    assert len(list(path(bexpr, (True, False, False, False)))) == 3
    assert len(list(path(bexpr, (True, True, True, True)))) == 5

    def merge(ctx, val, acc):
        if ctx.is_leaf:
            return ctx.path_negated ^ ctx.node_val
        return None

    def evaluate(vals):
        return fold_path(merge, bexpr, vals, initial=[])

    for val in product(*(4*[[False, True]])):
        expected = (val[0] == val[1]) and (val[2] == val[3])
        assert evaluate(val) == expected



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['fold_bdd']

package_data = \
{'': ['*']}

install_requires = \
['attrs>=19.3,<20.0', 'dd>=0.5.4,<0.6.0', 'funcy>=1.13,<2.0']

setup_kwargs = {
    'name': 'fold-bdd',
    'version': '0.4.0',
    'description': 'Library for folding (or reducing) over a Reduced Ordered Binary Decision Diagram.',
    'long_description': "# Fold-BDD\nLibrary for folding (or reducing) over a Reduced Ordered Binary Decision Diagram.\n\n[![Build Status](https://cloud.drone.io/api/badges/mvcisback/fold-bdd/status.svg)](https://cloud.drone.io/mvcisback/fold-bdd)\n[![codecov](https://codecov.io/gh/mvcisback/fold-bdd/branch/master/graph/badge.svg)](https://codecov.io/gh/mvcisback/fold-bdd)\n[![PyPI version](https://badge.fury.io/py/fold-bdd.svg)](https://badge.fury.io/py/fold-bdd)\n[![License: MIT](https://img.shields.io/badge/License-MIT-yellow.svg)](https://opensource.org/licenses/MIT)\n\n<!-- markdown-toc start - Don't edit this section. Run M-x markdown-toc-refresh-toc -->\n**Table of Contents**\n\n- [Fold-BDD](#fold-bdd)\n- [Installation](#installation)\n- [Usage](#usage)\n    - [Create ROBDD](#create-robdd)\n    - [Post-Order Examples](#post-order-examples)\n    - [Fold Path Examples](#fold-path-examples)\n- [Context Object Attributes](#context-object-attributes)\n\n<!-- markdown-toc end -->\n\n\n# Installation\n\nIf you just need to use `fold_bdd`, you can just run:\n\n`$ pip install fold-bdd`\n\nFor developers, note that this project uses the\n[poetry](https://poetry.eustace.io/) python package/dependency\nmanagement tool. Please familarize yourself with it and then\nrun:\n\n`$ poetry install`\n\n# Usage\n\nThe `fold-bdd` library supports two types of folds:\n\n1. Folding over the DAG of a `BDD` starting at the root and then\n   recursively merging the low and high branches until the\n   `True`/`False` leaves. This is simply a compressed variant\n   of a post-order traversal.\n\n2. Folding over a path in the DAG, starting at the root and moving the\n   the corresponding leaf (left fold).\n\nIn both cases, local context such as the levels of the parent and\nchild nodes are passed in.\n\nAs input, each of these take in a bdd, from the\n[dd](https://github.com/tulip-control/dd) library and function for\naccumulating or merging. \n\nThe following example illustrates how to use `fold_bdd` to count the\nnumber of solutions to a predicate using `post_order` and evaluate a\npath using `fold_path`.\n\n## Create ROBDD\n```python\n# Create BDD.\nfrom dd.cudd import BDD\n\nmanager = BDD()\nmanager.declare('x', 'y')\nmanager.reorder({'x': 1, 'y': 0})\nmanger.configure(reordering=False)\n\nbexpr = manager.add_expr('x | y')\n```\n\n## Post-Order Examples\n\n```python\nfrom fold_bdd import post_order\n```\n\n### Count Number of Nodes in BDD\n\n```python\ndef merge1(ctx, low=None, high=None):\n    return 1 if low is None else low + high\n\ndef dag_size(bexpr):\n    return post_order(bexpr, merge1)\n\nassert bexpr.dag_size == dag_size(bexpr)\n```\n\n### Count Number of Solutions to bexpr.\n```python\ndef merge2(ctx, low, high):\n    if ctx.is_leaf:\n        return ctx.skipped_paths if ctx.node_val else 0\n    return (low + high) * ctx.skipped_paths\n\ndef count_solutions(bexpr):\n    return post_order(bexpr, merge2)\n\nassert count_solutions(bexpr) == 3\n```\n\n## Fold Path Examples\n\n### Count nodes along path.\n\n```python\ndef merge(ctx, val, acc):\n    return acc + 1\n\ndef count_nodes(bexpr, vals):\n    return fold_path(merge, bexpr, vals, initial=0)\n\nassert count_nodes(bexpr, (False, False)) == 3\nassert count_nodes(bexpr, (True, False)) == 2\n```\n\n### Count paths corresponding to BDD path\n\n```python\ndef merge(ctx, val, acc):\n    return acc * ctx.skipped_paths\n\ndef count_paths(bexpr, vals):\n    return fold_path(merge, bexpr, vals, initial=1)\n\nassert count_paths(bexpr, (False, True)) == 1\nassert count_paths(bexpr, (True, False)) == 2\n```\n\n# Context Object Attributes\n\nThe `Context` object contains exposes attributes\n\n- `node: Hashable`  # Reference to Node in ROBDD.\n- `node_val: Union[str, bool]`  # Node name or leaf value.\n- `negated: bool`  # Is the edge to prev node negated.\n- `max_lvl: int`  # How many decision variables are there. \n- `curr_lvl: int`  # Which decision is this.\n- `prev_lvl: Optional[int]`  # Which decision was the parent. None if root.\n- `low_lvl: Optional[int]`  # Which decision does the False edge point to. None if leaf.\n- `high_lvl: Optional[int]`  # Which decision does the True edge point to. None if leaf.\n- `is_leaf: bool`  # Is the current node a leaf.\n- `skipped: int`  # How many decisions were skipped on edge to this node.\n",
    'author': 'Marcell Vazquez-Chanlatte',
    'author_email': 'mvc@linux.com',
    'url': 'https://github.com/mvcisback/fold-bdd',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
